import os
import sys
from . import BaseAction

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"


class ExecuteLocal(BaseAction):

    def __init__(self, run_cmd):

        # Need to expand users, get absolute path and dereference symlinks
        self.run_cmd = os.path.realpath(os.path.expanduser(run_cmd))

    def act_on_dir(self, dirname):

        full_cmd = 'cd ' + dirname + '\n' + self.run_cmd + '\n'
        result = os.system(full_cmd)
        if result != 0:
            sys.exit("Non-zero exit code from command '" + full_cmd + "'\n")



import logging
import chaospy as cp
from easyvvuq import OutputType
from .base import BaseAnalysisElement

# author: Jalal Lakhlili
__license__ = "LGPL"

logger = logging.getLogger(__name__)

# TODO:
# 1. Work out how to get multiple Sobol indices
# 1. a. Note that may require different orders for different qoi?
# 2. Add pd.read_hdf (https://pandas.pydata.org/pandas-docs/stable/user_guide/io.html#io-hdf5).
# 3. Test cp.fit_regression to approximate solver.


class PCEAnalysis(BaseAnalysisElement):

    def element_name(self):
        return "PCE_Analysis"

    def element_version(self):
        return "0.3"

    def __init__(self, params_cols=None, sampler=None, qoi_cols=None,
                 sobol_order=1):

        if sampler is None:
            msg = 'PCE analysis requires a paired sampler to be passed'
            raise RuntimeError(msg)

        if qoi_cols is None:
            raise RuntimeError("Analysis element requires a list of "
                               "quantities of interest (qoi)")

        if sobol_order > len(sampler.vary.vary_dict):
            logger.warning("sobol_order too high - set to number of "
                           "variables from sampler")
            sobol_order = len(sampler.vary.vary_dict)

        self.sobol_order = sobol_order
        self.params_cols = params_cols
        self.qoi_cols = qoi_cols
        self.output_type = OutputType.SUMMARY
        self.sampler = sampler

    def analyse(self, data_frame=None):

        if data_frame is None:
            raise RuntimeError("Analysis element needs a data frame to "
                               "analyse")

        qoi_cols = self.qoi_cols

        results = {'statistical_moments': {},
                   'percentiles': {},
                   'sobol_indices': {k: {} for k in qoi_cols},
                   'correlation_matrices': {},
                   'output_distributions': {},
                   }

        # Get the Polynomial
        P = self.sampler.P

        # Compute nodes and weights
        nodes, weights = cp.generate_quadrature(order=self.sampler.quad_order,
                                                domain=self.sampler.distribution,
                                                rule=self.sampler.quad_rule,
                                                sparse=self.sampler.quad_sparse)

        # Extract output values for each quantity of interest from Dataframe
        samples = {k: [] for k in qoi_cols}
        for run_id in data_frame.run_id.unique():
            for k in qoi_cols:
                values = data_frame.loc[data_frame['run_id'] == run_id][k]
                samples[k].append(values)

        output_distributions = {}

        # Compute descriptive statistics for each quantity of interest
        for k in qoi_cols:
            # Approximation solver
            fit = cp.fit_quadrature(P, nodes, weights, samples[k])

            # Statistical moments
            mean = cp.E(fit, self.sampler.distribution)
            var = cp.Var(fit, self.sampler.distribution)
            std = cp.Std(fit, self.sampler.distribution)
            results['statistical_moments'][k] = {'mean': mean,
                                                 'var': var,
                                                 'std': std}

            # Percentiles (Pxx)
            P10 = cp.Perc(fit, 10, self.sampler.distribution)
            P90 = cp.Perc(fit, 90, self.sampler.distribution)
            results['percentiles'][k] = {'p10': P10, 'p90': P90}

            # First Sobol indices
            # TODO: Implement higher order indices
            logger.warning('Only first order Sobol indices implemented')
            sobol_first_narr = cp.Sens_m(fit, self.sampler.distribution)
            sobol_first_dict = {}
            i_par = 0
            for param_name in self.sampler.vary.get_keys():
                sobol_first_dict[param_name] = sobol_first_narr[i_par]
                i_par += 1
            # 1 here = order
            results['sobol_indices'][k][1] = sobol_first_dict

            # Correlation matrix
            results['correlation_matrices'][k] = cp.Corr(
                fit, self.sampler.distribution)

            # Output distributions
            results['output_distributions'][k] = cp.QoI_Dist(
                fit, self.sampler.distribution)

        return results



from enum import Enum

# TODO: make this auto update with setup.py etc.
__easyvvuq_version__ = '0.3.dev1'

default_campaign_prefix = 'EasyVVUQ_Campaign_'


class OutputType(Enum):
    """Types of data output by UQPs/VVPs
    """

    SAMPLE = 'sample'
    ARRAY = 'array'
    SAMPLE_ARRAY = 'sample_array'
    TRACK = 'track'
    SUMMARY = 'summary'



from .base import BaseCollationElement
from easyvvuq import OutputType
import pandas as pd

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"


class AggregateSamples(BaseCollationElement, collater_name="aggregate_samples"):
    """
    Aggregate the results of all completed simulations described by the
    Campaign.

    Parameters
    ----------
    average:
        Should the values read in be averaged (mean).
    """

    def element_version(self):
        return "0.1"

    def __init__(self, average=False):
        self.average = average

    def collate(self, campaign):
        """
        Returns
        -------
        `pd.DataFrame`:
            Aggregated data from all completed runs referenced in the input Campaign.
        """
        decoder = campaign._active_app_decoder

        if decoder.output_type != OutputType.SAMPLE:
            raise RuntimeError('Can only aggregate sample type data')

        full_data = pd.DataFrame()

        # TODO: Find nicer way than forcing collate to access deep internal
        # vars of campaign object like this
        runs = campaign.campaign_db.runs()

        for run_id, run_info in runs.items():
            if decoder.sim_complete(run_info=run_info):

                campaign.campaign_db.set_run_status(run_id, "completed")

                run_data = decoder.parse_sim_output(run_info=run_info)

                if self.average:
                    run_data = pd.DataFrame(run_data.mean()).transpose()

                params = run_info['params']

                column_list = list(params.keys()) + run_data.columns.tolist()

                for param, value in params.items():
                    run_data[param] = value

                # Reorder columns
                run_data = run_data[column_list]
                run_data['run_id'] = run_id
                full_data = full_data.append(run_data, ignore_index=True)

        return full_data

    def get_restart_dict(self):
        return {"average": self.average}



import os
import pandas as pd
import numpy as np
import chaospy as cp
from itertools import product, chain, combinations
from easyvvuq import OutputType
from .base import BaseAnalysisElement


# Author: Wouter Edeling
__license__ = "LGPL"


class SCAnalysis(BaseAnalysisElement):

    def element_name(self):
        return "SC_analysis"

    def element_version(self):
        return "0.3"

    def __init__(self, params_cols=None, sampler=None, qoi_cols=None):

        if sampler is None:
            msg = 'PCE analysis requires a paired sampler to be passed'
            raise RuntimeError(msg)

        if qoi_cols is None:
            raise RuntimeError("Analysis element requires a list of "
                               "quantities of interest (qoi)")

        self.params_cols = params_cols
        self.qoi_cols = qoi_cols
        self.output_type = OutputType.SUMMARY
        self.sampler = sampler

    # main analysis subroutine
    def analyse(self, data_frame=None):

        if data_frame is None:
            raise RuntimeError("Analysis element needs a data frame to "
                               "analyse")

        qoi_cols = self.qoi_cols

        results = {'statistical_moments': {},
                   'sobol_indices': {k: {} for k in qoi_cols}}

        # Chaospy computation of 1D weights
        xi = []
        wi = []
        for dist in self.sampler.vary.get_values():
            xi_i, wi_i = cp.generate_quadrature(
                self.sampler.quad_order, dist, rule=self.sampler.quad_rule)
            xi.append(xi_i.flatten())
            wi.append(wi_i.flatten())
        self.xi = xi
        self.wi = wi

        # Compute tensor product nodes and weights
        xi_d, self.wi_d = cp.generate_quadrature(order=self.sampler.quad_order,
                                                 domain=self.sampler.joint_dist,
                                                 rule=self.sampler.quad_rule)
        self.xi_d = xi_d.T

        # Extract output values for each quantity of interest from Dataframe
        samples = {k: [] for k in qoi_cols}
        for run_id in data_frame.run_id.unique():
            for k in qoi_cols:
                values = data_frame.loc[data_frame['run_id'] == run_id][k]
                samples[k].append(values)

        self.samples = samples
        self._number_of_samples = self.sampler._number_of_samples

        # number of uncertain parameters
        self.d = self.xi_d.shape[1]

        # size of one code sample
        self.N_qoi = self.samples[qoi_cols[0]][0].size

        # Compute descriptive statistics for each quantity of interest
        for qoi_k in qoi_cols:
            mean_k, var_k = self.get_moments(qoi_k)
            std_k = var_k**0.5

            # compute statistical moments
            results['statistical_moments'][qoi_k] = {'mean': mean_k,
                                                     'var': var_k,
                                                     'std': std_k}

            # compute all Sobol indices
            results['sobol_indices'][qoi_k] = self.get_Sobol_indices(qoi_k, 'all')

        return results

    # Compute the first two statistical moments
    def get_moments(self, qoi):

        # compute the mean and variance of the code samples, using quad weights
        mean_f = np.zeros([self.N_qoi, 1])
        var_f = np.zeros([self.N_qoi, 1])

        for k in range(self._number_of_samples):
            sample_k = (self.samples[qoi][k]).values.reshape([self.N_qoi, 1])
            mean_f += sample_k * self.wi_d[k].prod()

        for k in range(self._number_of_samples):
            sample_k = (self.samples[qoi][k]).values.reshape([self.N_qoi, 1])
            var_f += (sample_k - mean_f)**2 * self.wi_d[k].prod()

        return mean_f, var_f

    # use the SC expansion as a surrogate

    def surrogate(self, qoi, x):
        # interpolated QoI
        f_int = np.zeros([self.N_qoi, 1])

        # list with the 1d collocation points of all uncertain parameters
        # [self.all_vars[param]['xi_1d'] for param in self.all_vars.keys()]
        C = self.xi

        # loop over all samples
        for k in range(self._number_of_samples):

            idx = {}
            for i in range(self.d):
                # indices of current collocation point xi_d[k] in 1d
                # collocation points
                idx[i] = (C[i] == self.xi_d[k][i]).nonzero()[0]

            L = []
            for i in range(self.d):
                # values of Lagrange polynomials at x
                L.append(LagrangePoly(x[i], C[i], idx[i]))

            # current sample
            qoi_k = self.samples[qoi][k].values.reshape([self.N_qoi, 1])  # .reshape(self.N_qoi)

            # surrogate: samples interpolated via Lagrange polynomials
            f_int += qoi_k * np.prod(L)

        return f_int

    ####################################################################
    # BEGIN SOBOL SUBROUTINES                                          #
    # Method: 'Global Sensitivity Analysis for Stocastic Collocation'  #
    #          G. Tang and G. Iaccarino, AIAA 2922, 2010               #
    ####################################################################

    def compute_tensor_prod_u(self, xi, wi, u, u_prime):
        # tensor products with dimension of u
        xi_u = {}
        wi_u = {}
        for key in u:
            xi_u[key] = xi[key]
            wi_u[key] = wi[key]

        xi_d_u = np.array(list(product(*xi_u.values())))
        wi_d_u = np.array(list(product(*wi_u.values())))

        # tensor products with dimension of u' (complement of u)
        xi_u_prime = {}
        wi_u_prime = {}
        for key in u_prime:
            xi_u_prime[key] = xi[key]
            wi_u_prime[key] = wi[key]

        xi_d_u_prime = np.array(list(product(*xi_u_prime.values())))
        wi_d_u_prime = np.array(list(product(*wi_u_prime.values())))

        return {'xi_d_u': xi_d_u, 'wi_d_u': wi_d_u,
                'xi_d_u_prime': xi_d_u_prime, 'wi_d_u_prime': wi_d_u_prime}

    #############################

    def compute_h(self, qoi, u, u_prime, xi_d_u, xi_d_u_prime, wi_d_u_prime):
        S_u = xi_d_u.shape[0]
        S_u_prime = xi_d_u_prime.shape[0]

        # coefficients h = f*w' integrated over u', so cardinality is that of u
        h = {}
        for i_u in range(S_u):
            h[i_u] = 0.
            for i_up in range(S_u_prime):

                # collocation point to be evaluated
                xi_s = np.zeros(self.d)

                # add the xi of u (at the correct location k)
                idx = 0
                for k in u:
                    xi_s[k] = xi_d_u[i_u][idx]
                    idx += 1

                # add the xi of u' (at the correct location k)
                idx = 0
                for k in u_prime:
                    xi_s[k] = xi_d_u_prime[i_up][idx]
                    idx += 1

                #
                tmp = np.prod(self.xi_d == xi_s, axis=1)
                idx = np.where(tmp == 1)[0][0]
                h[i_u] += self.samples[qoi][idx].values.flatten() * \
                    wi_d_u_prime[i_up].prod()

        return h

    #############################

    # Computes Sobol indices using Stochastic Collocation
    def get_Sobol_indices(self, qoi, typ):

        # multi indices
        U = list(range(self.d))

        if typ == 'first_order':
            P = list(powerset(U))[0:self.d + 1]
        elif typ == 'all':
            # all indices u
            P = list(powerset(U))

        # get first two moments
        mu, D = self.get_moments(qoi)
        mu = mu.flatten()
        D = D.flatten()

        # list of 1D nodes and quad weights
        xi = self.xi
        wi = self.wi

        # total variance might be zero at some locations, Sobol index not defined there
        # idx_gt0 = np.where(D > 0)[0]

        # partial variances
        D_u = {}
        # D_0 = mu**2
        D_u[P[0]] = mu**2

        sobol = {}

        for u in P[1:]:

            # complement of u
            u_prime = np.delete(U, u)

            # compute corresponding tensor products and GQ weights
            tmp = self.compute_tensor_prod_u(xi, wi, u, u_prime)
            xi_d_u = tmp['xi_d_u']
            wi_d_u = tmp['wi_d_u']
            xi_d_u_prime = tmp['xi_d_u_prime']
            wi_d_u_prime = tmp['wi_d_u_prime']

            # cardinality of u
            S_u = xi_d_u.shape[0]

            # h coefficients
            h = self.compute_h(qoi, u, u_prime, xi_d_u, xi_d_u_prime, wi_d_u_prime)

            # partial variance
            D_u[u] = 0.0
            for i_u in range(S_u):
                D_u[u] += h[i_u]**2 * wi_d_u[i_u].prod()
            D_u[u] = D_u[u].flatten()

            # all subsets of u
            W = list(powerset(u))[0:-1]

            # partial variance of u
            for w in W:
                D_u[u] -= D_u[w]

            # compute Sobol index, only include points where D > 0
            # sobol[u] = D_u[u][idx_gt0]/D[idx_gt0]
            sobol[u] = D_u[u] / D

        sort = []
        for u in P[1:]:
            #print('Sobol index ', u, ' = ', sobol[u])
            sort.append(sobol[u])

        # print('Total sum = ', np.sum(sobol.values())/self.N_qoi)

        return sobol

    #########################
    # END SOBOL SUBROUTINES #
    #########################

# https://docs.python.org/3/library/itertools.html#recipes


def powerset(iterable):
    "powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(len(s) + 1))

# Lagrange polynomials used for interpolation


def LagrangePoly(x, x_i, j):

    l_j = 1.0

    for i in range(len(x_i)):

        if i != j:
            denom = x_i[j] - x_i[i]
            nom = x - x_i[i]

            l_j *= nom / denom

    return l_j
